"""
Argument parser
"""
from typing import Sequence, Dict, List
import os.path
import argparse
# pylint: disable=line-too-long

from .utility import SUPPORTED_FORMATS
from .constants import CASTEP_OUTPUT_NAMES, CASTEP_FILE_FORMATS

AP = argparse.ArgumentParser(
    prog="castep_outputs",
    description=f"""Attempts to find all files for seedname, filtered by `inc` args (default: all).
    Explicit files can be passed using longname arguments.
    castep_outputs can parse most human-readable castep outputs including: {', '.join(CASTEP_FILE_FORMATS)}"""
)

AP.add_argument("seedname", nargs=argparse.REMAINDER, help="Seed name for data")
AP.add_argument("-V", "--version", action="version", version="%(prog)s v0.1")
AP.add_argument("-L", "--log", help="Verbose output",
                choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'), default="WARNING")
AP.add_argument("-o", "--output", help="File to write output, default: screen", default=None)
AP.add_argument("-f", "--out-format", help="Output format", choices=SUPPORTED_FORMATS, default="json")

AP.add_argument("-t", "--testing", action="store_true", help="Set testing mode to produce flat outputs")

AP.add_argument("-A", "--inc-all", action="store_true", help="Extract all available information")
AP.add_argument("-c", "--inc-castep", action="store_true", help="Extract .castep information")
AP.add_argument("-g", "--inc-geom", action="store_true", help="Extract .geom information")
AP.add_argument("-m", "--inc-md", action="store_true", help="Extract .md information")
AP.add_argument("-b", "--inc-bands", action="store_true", help="Extract .bands information")
AP.add_argument("-p", "--inc-phonon_dos", action="store_true", help="Extract .phonon_dos information")
AP.add_argument("-e", "--inc-efield", action="store_true", help="Extract .efield information")
AP.add_argument("-x", "--inc-xrd_sf", action="store_true", help="Extract .xrd_sf information")
AP.add_argument("-H", "--inc-hug", action="store_true", help="Extract .hug information")
AP.add_argument("-E", "--inc-elf_fmt", action="store_true", help="Extract .elf_fmt information")
AP.add_argument("-C", "--inc-chdiff_fmt", action="store_true", help="Extract .chdiff_fmt information")
AP.add_argument("-P", "--inc-pot_fmt", action="store_true", help="Extract .pot_fmt information")
AP.add_argument("-D", "--inc-den_fmt", action="store_true", help="Extract .den_fmt information")
AP.add_argument("-X", "--inc-elastic", action="store_true", help="Extract .elastic information")
AP.add_argument("-T", "--inc-ts", action="store_true", help="Extract .ts information")

AP.add_argument('--inc-param', action="store_true", help="Extract .param information")
AP.add_argument('--inc-cell', action="store_true", help="Extract .cell information")

AP.add_argument("--castep", nargs="*", help="Extract from CASTEP as .castep type", default=[])
AP.add_argument("--geom", nargs="*", help="Extract from GEOM as .geom type", default=[])
AP.add_argument("--cell", nargs="*", help="Extract from CELL as .cell type", default=[])
AP.add_argument("--param", nargs="*", help="Extract from PARAM as .param type", default=[])
AP.add_argument("--md", nargs="*", help="Extract from MD as .md type", default=[])
AP.add_argument("--bands", nargs="*", help="Extract from BANDS as .bands type", default=[])
AP.add_argument("--hug", nargs="*", help="Extract from HUG as .hug type", default=[])
AP.add_argument("--phonon_dos", nargs="*", help="Extract from PHONON_DOS as .phonon_dos type", default=[])
AP.add_argument("--efield", nargs="*", help="Extract from EFIELD as .efield type", default=[])
AP.add_argument("--xrd_sf", nargs="*", help="Extract from XRD_SF as .xrd_sf type", default=[])
AP.add_argument("--elf_fmt", nargs="*", help="Extract from ELF_FMT as .elf_fmt type", default=[])
AP.add_argument("--chdiff_fmt", nargs="*", help="Extract from CHDIFF_FMT as .chdiff_fmt type", default=[])
AP.add_argument("--pot_fmt", nargs="*", help="Extract from POT_FMT as .pot_fmt type", default=[])
AP.add_argument("--den_fmt", nargs="*", help="Extract from DEN_FMT as .den_fmt type", default=[])
AP.add_argument("--elastic", nargs="*", help="Extract from ELASTIC as .elastic type", default=[])
AP.add_argument("--ts", nargs="*", help="Extract from TS as .ts type", default=[])


def parse_args(to_parse: Sequence[str] = ()) -> argparse.Namespace:
    """ Parse all arguments and add those caught by flags """
    args = AP.parse_args()

    parse_all = args.inc_all or not any(getattr(args, f"inc_{typ}") for typ in CASTEP_OUTPUT_NAMES)

    # Set all flags
    if parse_all and not to_parse:
        for typ in CASTEP_OUTPUT_NAMES:
            setattr(args, f"inc_{typ}", True)

    # Only parse those which are requested
    for typ in to_parse:
        setattr(args, f"inc_{typ}", True)

    # Add seeded files into parse list if to be included
    for seed in args.seedname:
        if os.path.isfile(seed) and (ext := os.path.splitext(seed)[1][1:]) in CASTEP_OUTPUT_NAMES:
            getattr(args, ext).append(seed)
        else:
            for typ in CASTEP_OUTPUT_NAMES:
                if getattr(args, f"inc_{typ}") and os.path.isfile(seed+typ):
                    getattr(args, typ).append(seed+typ)

    return args


def args_to_dict(args: argparse.Namespace) -> Dict[str, List[str]]:
    """ Convert args namespace to dictionary """
    out_dict = {typ: getattr(args, typ) for typ in CASTEP_OUTPUT_NAMES}
    return out_dict



"""
Module to parse miscellaneous castep files
"""

__AUTHOR__ = "Jacob Wilkins"
__VERSION__ = "0.1.4"

# pylint: disable=unused-import

from .castep_outputs_main import parse_single
from .castep_file_parser import parse_castep_file
from .cell_param_file_parser import parse_cell_param_file, parse_cell_file, parse_param_file
from .md_geom_file_parser import parse_md_geom_file, parse_md_file, parse_geom_file
from .extra_files_parser import (parse_bands_file, parse_hug_file, parse_phonon_dos_file,
                                 parse_efield_file, parse_xrd_sf_file, parse_elf_fmt_file,
                                 parse_chdiff_fmt_file, parse_pot_fmt_file, parse_den_fmt_file,
                                 parse_elastic_file, parse_ts_file)



"""
Parse castep .md or .geom files
"""

from typing import TextIO, List, Dict
from collections import defaultdict

from .castep_res import ATDATTAG, TAG_RE, get_numbers
from .constants import FST_D, TAG_ALIASES
from .utility import to_type, add_aliases, atreg_to_index


def parse_md_geom_file(md_geom_file: TextIO) -> List[Dict[str, float]]:
    """ Parse standard MD and GEOM file formats """

    while "END header" not in md_geom_file.readline():
        pass

    steps = []
    curr = defaultdict(list)
    for line in md_geom_file:
        if not line.strip():  # Next step
            if curr:
                add_aliases(curr, TAG_ALIASES)
                for ion in filter(lambda x: isinstance(x, tuple), curr):
                    add_aliases(curr[ion], TAG_ALIASES)
                steps.append(curr)
            curr = defaultdict(list)
        elif not TAG_RE.search(line):  # Timestep
            curr['time'] = to_type(get_numbers(line)[0], float)

        elif match := ATDATTAG.match(line):
            ion = atreg_to_index(match)
            if ion not in curr:
                curr[ion] = {}
            curr[ion][match.group('tag')] = to_type([*(match.group(d) for d in FST_D)], float)

        elif match := TAG_RE.search(line):
            curr[match.group('tag')].append([*to_type(get_numbers(line), float)])

    return steps


# Aliases
parse_md_file = parse_md_geom_file
parse_geom_file = parse_md_geom_file



"""
Utility functions for parsing castep outputs
"""
from typing import Tuple, Any, Dict, TextIO, List, Callable, Union
import fileinput
import logging
import collections.abc
from collections import defaultdict
import json
import pprint
try:
    from ruamel import yaml
    _YAML_TYPE = "ruamel"
except ImportError:
    try:
        import yaml
        _YAML_TYPE = "yaml"
    except ImportError:
        _YAML_TYPE = None


def normalise_string(string: str) -> str:
    """
    Normalise a string removing leading/trailing space
    and making all spacing single-space
    """
    return " ".join(string.strip().split())


def atreg_to_index(dict_in: dict, clear: bool = True) -> Tuple[str, int]:
    """
    Transform a matched atreg value to species index tuple
    Also clear value from dictionary for easier processing
    """
    spec, ind = dict_in["spec"], dict_in["index"]

    if isinstance(dict_in, dict) and clear:
        del dict_in["spec"]
        del dict_in["index"]

    return (spec, int(ind))


def normalise(obj: Any) -> Any:
    """
    Standardises data after processing
    """
    if isinstance(obj, (tuple, list)):
        obj = tuple(normalise(v) for v in obj)
    elif isinstance(obj, (dict, defaultdict)):
        obj = {key: normalise(val) for key, val in obj.items()}

    return obj


def json_safe(obj: Any) -> Any:
    """ Transform datatypes into JSON safe variants"""
    if isinstance(obj, (tuple, list)):
        obj = [json_safe(v) for v in obj]
    elif isinstance(obj, dict):
        obj = json_safe_dict(obj)
    elif isinstance(obj, complex):
        obj = (obj.real, obj.imag)
    return obj


def json_safe_dict(obj: Dict) -> Dict:
    """ Transform a castep_output dict into a JSON safe variant
    i.e. convert tuple keys to conjoined strings """
    obj_out = {}

    for key, val in obj.items():
        if isinstance(key, (tuple, list)):
            key = "_".join(map(str, key))
        obj_out[key] = json_safe(val)
    return obj_out


def flatten_dict(dictionary: Dict,
                 parent_key: bool = False,
                 separator: str = '_') -> Dict[str, Any]:
    """
    Turn a nested dictionary into a flattened dictionary

    Taken from:
    https://stackoverflow.com/a/62186053

    :param dictionary: The dictionary to flatten
    :param parent_key: The string to prepend to dictionary's keys
    :param separator: The string used to separate flattened keys
    :return: A flattened dictionary
    """

    items = []
    for key, value in dictionary.items():
        new_key = str(parent_key) + separator + key if parent_key else key
        if isinstance(value, collections.abc.MutableMapping):
            items.extend(flatten_dict(value, new_key, separator).items())
        elif isinstance(value, list):
            for keyx, val in enumerate(value):
                items.extend(flatten_dict({str(keyx): val}, new_key).items())
        else:
            items.append((new_key, value))
    return dict(items)


def json_dumper(data: Any, file: TextIO):
    """ Basic JSON format dumper """
    json.dump(data, file, indent=2)


def ruamel_dumper(data: Any, file: TextIO):
    """ Basic ruamel.yaml format dumper """
    yaml_eng = yaml.YAML(typ='unsafe')
    yaml_eng.dump(data, file)


def yaml_dumper(data: Any, file: TextIO):
    """ Basic yaml format dumper """
    yaml.dump(data, file)


def pprint_dumper(data: Any, file: TextIO):
    """ PPrint dumper """
    print(pprint.pformat(data), file=file)


def print_dumper(data: Any, file: TextIO):
    """ Print dumper """
    print(data, file=file)


SUPPORTED_FORMATS = {"json": json_dumper,
                     "ruamel": ruamel_dumper,
                     "yaml": yaml_dumper,
                     "pprint": pprint_dumper,
                     "print": print_dumper}


def get_dumpers(dump_fmt: str) -> Callable:
    """
    Get appropriate dump for unified interface
    """
    if dump_fmt not in SUPPORTED_FORMATS:
        raise ValueError(f"Cannot output in {dump_fmt} format.")

    if dump_fmt == "yaml":
        if _YAML_TYPE is None:
            raise ImportError("Couldn't find valid yaml dumper (ruamel.yaml / yaml)"
                              "please install and try again.")
        dump_fmt = _YAML_TYPE

    return SUPPORTED_FORMATS[dump_fmt]


def stack_dict(out_dict: Dict[Any, List], in_dict: Dict[Any, List]) -> Dict[Any, List]:
    """ Append items in `in_dict` to the keys in `out_dict` """
    for key, val in in_dict.items():
        out_dict[key].append(val)


def fix_data_types(in_dict: Dict, type_dict: Dict[str, type]) -> Dict:
    """ Applies correct types to elements of in_dict by mapping given in type_dict"""
    for key, typ in type_dict.items():
        if key in in_dict:
            in_dict[key] = to_type(in_dict[key], typ)


def add_aliases(in_dict: Dict[str, Any],
                alias_dict: [str, str],
                replace: bool = False) -> Dict[str, Any]:
    """ Adds aliases of known names into dictionary, if replace is true, remove original """
    for frm, new in alias_dict.items():
        if frm in in_dict:
            in_dict[new] = in_dict[frm]
            if replace:
                in_dict.pop(frm)


def to_type(data_in: Union[str, List, Tuple], typ: type) -> Union[Tuple[type], type]:
    """ Convert types to `typ` regardless of if data_in is iterable or otherwise """
    if isinstance(data_in, (list, tuple)):
        data_in = tuple(map(typ, data_in))
    elif isinstance(data_in, str):
        data_in = typ(data_in)
    return data_in


def log_factory(file: TextIO) -> Callable:
    """ Return logging function to add file info to logs """
    if isinstance(file, fileinput.FileInput):
        def log_file(message, *args, level="info"):
            getattr(logging, level)(f"[{file.filename()}:{file.lineno()}]"
                                    f" {message}", *args)
    elif hasattr(file, 'name'):
        def log_file(message, *args, level="info"):
            getattr(logging, level)(f"[{file.name}] {message}", *args)
    else:
        def log_file(message, *args, level="info"):
            getattr(logging, level)(message, *args)

    return log_file



""" Module containing all regexes """

from typing import List, Sequence, Optional, TextIO
import re
import io
import itertools

from .constants import MINIMISERS, SHELLS, FST_D


def get_numbers(line: str) -> List[str]:
    """ Get all numbers in a string as a list """
    return NUMBER_RE.findall(line)


def get_block(init_line: str, in_file: TextIO,
              start: re.Pattern, end: re.Pattern, *, cnt: int = 1,
              out_fmt: type = io.StringIO, eof_possible: bool = False):
    """ Check if line is the start of a block and return
    the block if it is, moving in_file forward as it does so """

    block = ""

    if not re.search(start, init_line):
        return block

    block = init_line
    fnd = cnt
    for line in in_file:
        block += line
        if re.search(end, line):
            fnd -= 1
            if fnd == 0:
                break
    else:
        if not eof_possible:
            if hasattr(in_file, 'name'):
                raise IOError(f"Unexpected end of file in {in_file.name}.")
            raise IOError("Unexpected end of file.")

    if not block:
        return ""
    if out_fmt is str:
        return block
    if out_fmt is list:
        return block.splitlines()
    if out_fmt is io.StringIO:
        return io.StringIO(block)
    return out_fmt(block)


def labelled_floats(labels: Sequence[str], counts: Sequence[Optional[int]] = (None,),
                    sep: str = r"\s+?", suff: str = "") -> str:
    """ Constructs a regex for extracting floats with assigned labels
    :param labels:iterable of labels to label each group
    :param counts:iterable of counts to group into each label (count must not exceed that of labels)
    :param sep:separator between floats
    """
    if suff and any(cnt for cnt in counts):
        raise NotImplementedError("Suffix and counts not currently supported")

    outstr = ""
    for label, cnt in itertools.zip_longest(labels, counts):
        if cnt:
            outstr += f"(?:(?P<{label}>(?:{sep}{EXPNUMBER_RE}{suff}){{{cnt}}}))"
        else:
            outstr += f"(?:{sep}(?P<{label}>{EXPNUMBER_RE}){suff})"

    return outstr


def gen_table_re(content: str, border: str = r"\s*",
                 *, pre: str = "", post: str = "", whole_line: bool = True):
    """ Constructs a regex for matching table headers with given borders """

    tab_re = (rf"\s*{border}\s*{content}\s*{border}\s*"
              if content else
              rf"\s*{border}\s*")

    tab_re = rf"\s*{pre}{tab_re}{post}\s*"

    if whole_line:
        tab_re = f"^{tab_re}$"

    return tab_re


# --- RegExes
# Regexps to recognise numbers
FNUMBER_RE = r"(?:[+-]?(?:\d*\.?\d+|\d+\.?\d*))"
EXPNUMBER_RE = rf"(?:{FNUMBER_RE}(?:[Ee][+-]?\d{{1,3}})?)"
INTNUMBER_RE = r"(?:\d+)"
NUMBER_RE = re.compile(rf"(?:{EXPNUMBER_RE}|{FNUMBER_RE}|{INTNUMBER_RE})")

# Regexp to identify extended chemical species
SPECIES_RE = r"[A-Z][a-z]{0,2}"
ATOM_NAME_RE = rf"\b{SPECIES_RE}(?::\w+)?\b"

# Unless we have *VERY* exotic electron shells
SHELL_RE = rf"\d[{''.join(SHELLS)}]\d{{0,2}}"

TAG_RE = re.compile(r"<--\s*(?P<tag>\w+)")

EMPTY = r"^\s*$"

# Atom regexp
ATREG = rf"(?P<spec>{ATOM_NAME_RE})\s+(?P<index>\d+)"


# Atom reference with 3-vector
ATDAT3VEC = re.compile(ATREG + labelled_floats(FST_D))
ATDATTAG = re.compile(rf"\s*{ATDAT3VEC.pattern}\s*{TAG_RE.pattern}")

# SCF Loop
SCF_LOOP_RE = re.compile(r"\s*(?:Initial|\d+)\s*"
                         rf"{labelled_floats(('energy', 'fermi_energy', 'energy_gain'))}?\s*"
                         f"{labelled_floats(('time',))}")

# PS Energy
PS_SHELL_RE = re.compile(
    rf"\s*Pseudo atomic calculation performed for (?P<spec>{SPECIES_RE})(\s+{SHELL_RE})+")

# PS Projector
PSPOT_PROJ_RE = re.compile(r"(?P<orbital>\d)(?P<shell>\d)(?P<type>U|UU|N)?")
UNLABELLED_PROJ_RE = r"\d\d(?:UU|U|N)?"

PSPOT_REFERENCE_STRUC_RE = re.compile(
    rf"""
    ^\s*\|\s*
    (?P<orb>{SHELL_RE}(?:/\d+)?)\s*
    {labelled_floats(('occupation', 'energy'))}
    \s*\|\s*$
    """, re.VERBOSE)
PSPOT_DEF_RE = re.compile(
    rf"""
    ^\s*\|\s*
    (?P<beta>\d+|loc)\s*
    (?P<l>\d+)\s*
    (?P<j>\d+)?\s*
    {labelled_floats(('e', 'Rc'))}\s*
    (?P<scheme>\w+)\s*
    (?P<norm>\d+)
    \s*\|\s*$
    """, re.VERBOSE)

# PSPot String
PSPOT_RE = re.compile(labelled_floats(("local_channel",
                                       "core_radius",
                                       "beta_radius",
                                       "r_inner",
                                       "coarse",
                                       "medium",
                                       "fine"), sep=r"\|?")
                      +
                      r"\|"
                      rf"(?P<proj>{UNLABELLED_PROJ_RE}(?::{UNLABELLED_PROJ_RE})*)"
                      rf"(?:\{{(?P<shell_swp>{SHELL_RE}(?:,{SHELL_RE})*)\}})?"
                      rf"\((?P<opt>[^)]+)\)"
                      rf"(?P<debug>#)?"
                      rf"(?:\[(?P<shell_swp2>{SHELL_RE}(?:,{SHELL_RE})*)\])?"
                      )

# Forces block
FORCES_BLOCK_RE = re.compile(gen_table_re("([a-zA-Z ]*)Forces", r"\*+"), re.IGNORECASE)
# Stresses block
STRESSES_BLOCK_RE = re.compile(gen_table_re("([a-zA-Z ]*)Stress Tensor", r"\*+"), re.IGNORECASE)

# Bonds
BOND_RE = re.compile(rf"""\s*
                       (?P<spec1>{ATOM_NAME_RE})\s*(?P<ind1>\d+)\s*
                       --\s*
                       (?P<spec2>{ATOM_NAME_RE})\s*(?P<ind2>\d+)\s*
                       {labelled_floats(("population", "length"))}
                       """, re.VERBOSE)

# Pair pot
PAIR_POT_RES = {
    'two_body_one_spec': re.compile(
        rf"^(?P<tag>\w+)?\s*\*\s*(?P<spec>{ATOM_NAME_RE})\s*\*\s*$"
    ),
    'two_body_spec':  re.compile(
        rf"(?P<spec1>{ATOM_NAME_RE})\s*-\s*"
        rf"(?P<spec2>{ATOM_NAME_RE})"
    ),
    'two_body_val': re.compile(
        rf"""
            (?P<tag>\w+)?\s*\*\s*
            (?P<label>\w+)\s*
            {labelled_floats(('params',), counts=('1,4',))}\s*
            [\w^/*]+\s* \* \s*
            <--\s*(?P<type>\w+)
            """, re.ASCII | re.VERBOSE
    ),
    'three_body_spec': re.compile(
        rf"""
        ^(?P<tag>\w+)?\s*\*\s*
        (?P<spec>(?:{ATOM_NAME_RE}\s*){{3}})
        \s*\*\s*$""", re.VERBOSE
    ),
    'three_body_val': re.compile(
        rf"""
        ^(?P<tag>\w+)?\s*\*\s*
        (?P<label>\w+)\s*
        {labelled_floats(('params',))}\s*
        [\w^/*]+\s* \* \s*
        <--\s*(?P<type>\w+)
        """, re.VERBOSE
    )
}

# Orbital population
ORBITAL_POPN_RE = re.compile(rf"\s*{ATREG}\s*(?P<orb>[SPDF][xyz]?)"
                             rf"\s*{labelled_floats(('charge',))}")

# Regexp to identify phonon block in .castep file
PHONON_RE = re.compile(
    rf"""
    \s+\+\s+
    q-pt=\s*{INTNUMBER_RE}\s+
    \({labelled_floats(("qpt",), counts=(3,))}\)
    \s+
    ({FNUMBER_RE})\s+\+
    """, re.VERBOSE)

PROCESS_PHONON_RE = re.compile(
    rf"""\s+\+\s+
    (?P<N>\d+)\s+
    (?P<frequency>{FNUMBER_RE})\s*
    (?P<irrep>[a-zA-V])?\s*
    (?P<intensity>{FNUMBER_RE})?\s*
    (?P<active>[YN])?\s*
    (?P<raman_intensity>{FNUMBER_RE})?\s*
    (?P<raman_active>[YN])?\s*\+""", re.VERBOSE)

TDDFT_RE = re.compile(
    rf"""\s*\+\s*
    {INTNUMBER_RE}
    {labelled_floats(("energy", "error"))}
    \s*(?P<type>\w+)
    \s*\+TDDFT""", re.VERBOSE)

BS_RE = re.compile(
    rf"""
    Spin=\s*(?P<spin>{INTNUMBER_RE})\s*
    kpt=\s*{INTNUMBER_RE}\s*
    \({labelled_floats(("kx","ky","kz"))}\)\s*
    kpt-group=\s*(?P<kpgrp>{INTNUMBER_RE})
    """, re.VERBOSE)

THERMODYNAMICS_DATA_RE = re.compile(labelled_floats(("T", "E", "F", "S", "Cv")))

MINIMISERS_RE = f"(?:{'|'.join(map(lambda x: x.upper(), MINIMISERS))})"
GEOMOPT_MIN_TABLE_RE = re.compile(
    r"\s*\|\s* (?P<step>[^|]+)" +
    labelled_floats(("lambda", "Fdelta", "enthalpy"), sep=r"\s*\|\s*") +
    r"\s* \|", re.VERBOSE)

GEOMOPT_TABLE_RE = re.compile(
    r"\s*\|\s* (?P<parameter>\S+)" +
    labelled_floats(('value', 'tolerance'), sep=r"\s*\|\s*") +
    r"\s*\|\s* \S+ (?#Units) \s*\|\s* (?P<converged>No|Yes) \s*\|", re.VERBOSE)


# Regexp to identify Mulliken ppoulation analysis line
POPN_RE = re.compile(rf"\s*{ATREG}\s*(?P<spin_sep>up:)?" +
                     labelled_floats((*SHELLS, "total", "charge", "spin")) +
                     "?"   # Spin is optional
                     )

POPN_RE_DN = re.compile(r"\s+\d+\s*dn:" +
                        labelled_floats((*SHELLS, "total"))
                        )

# Regexp for born charges
BORN_RE = re.compile(rf"\s+{ATREG}(?P<charges>(?:\s*{FNUMBER_RE}){{3}})")

# MagRes REs
MAGRES_RE = (
    # "Chemical Shielding Tensor" 0
    re.compile(rf"\s*\|\s*{ATREG}{labelled_floats(('iso','aniso'))}\s*"
               rf"(?P<asym>{FNUMBER_RE}|N/A)\s*\|\s*"),
    # "Chemical Shielding and Electric Field Gradient Tensor" 1
    re.compile(rf"\s*\|\s*{ATREG}{labelled_floats(('iso','aniso'))}\s*"
               rf"(?P<asym>{FNUMBER_RE}|N/A)"
               rf"{labelled_floats(('cq', 'eta'))}\s*\|\s*"),
    # "Electric Field Gradient Tensor" 2
    re.compile(rf"\s*\|\s*{ATREG}{labelled_floats(('cq',))}\s*"
               rf"(?P<asym>{FNUMBER_RE}|N/A)\s*\|\s*"),
    # "(?:I|Ani)sotropic J-coupling" 3
    re.compile(rf"\s*\|\**\s*{ATREG}{labelled_floats(('fc','sd','para','dia','tot'))}\s*\|\s*"),
    # "Hyperfine Tensor" 4
    re.compile(rf"\s*\|\s*{ATREG}{labelled_floats(('iso',))}\s*\|\s*")
)

# MagRes Tasks
MAGRES_TASK = (
    "Chemical Shielding",
    "Chemical Shielding and Electric Field Gradient",
    "Electric Field Gradient",
    "(An)Isotropic J-coupling",
    "Hyperfine"
)

# Regexp to identify block in .phonon or .phonon_dos file
FRACCOORDS_RE = re.compile(rf"\s*(?P<index>{INTNUMBER_RE}){labelled_floats(('u', 'v', 'w'))}"
                           rf"\s*(?P<spec>{SPECIES_RE}){labelled_floats(('mass',))}")

PHONON_PHONON_RE = re.compile(rf"""
    \s+q-pt=\s*{INTNUMBER_RE}\s*
    {labelled_floats(('qpt', 'pth'), counts=(3, 1))}
    """, re.VERBOSE)

PROCESS_PHONON_PHONON_RE = re.compile(labelled_floats(('n', 'f', 'Grad_qf')))


# Regexp to identify Fermi energies in .bands file
BANDS_FERMI_RE = re.compile(r"Fermi energ(ies|y) \(in atomic units\)\s*" +
                            labelled_floats(('a', 'b')))

# Regexp to identify eigenvalue block in .bands file
# BANDS_EIGENS_RE =
# rf"K-point\s+(\d+)\s*(\s*{FNUMBER_RE})\s*({FNUMBER_RE})\s*({FNUMBER_RE})\s*({FNUMBER_RE})"

DEVEL_CODE_VAL_RE = r'[A-Za-z0-9_]+[:=]\S+'
DEVEL_CODE_BLOCK_RE = rf'([A-Za-z0-9_]+):(?:\s*{DEVEL_CODE_VAL_RE}\s*)*:end\S+'



"""
Parse the following castep outputs:
.bands

"""
from typing import TextIO, Sequence, Dict, Union, List, Any, Tuple
import re
from collections import defaultdict

from . import castep_res as REs
from .castep_res import get_block, get_numbers, labelled_floats, ATDATTAG, TAG_RE
from .constants import FST_D, SND_D, TAG_ALIASES, TS_TYPES
from .utility import (fix_data_types, stack_dict, to_type, log_factory, atreg_to_index,
                      add_aliases)


def parse_regular_header(block: TextIO,
                         extra_opts: Sequence[str] = tuple()) -> Dict[str, Union[float, int]]:
    """ Parse (semi-)standard castep file header block (given as iterable over lines) """

    data = {}
    coords = defaultdict(list)
    for line in block:

        if line.strip().startswith("Number of"):
            _, _, *key, val = line.split()
            data[" ".join(key)] = int(float(val))
        elif "Unit cell vectors" in line:
            data['unit_cell'] = [to_type(next(block).split(), float)
                                 for _ in range(3)]

        elif match := REs.FRACCOORDS_RE.match(line):
            stack_dict(coords, match.groupdict())

        elif match := re.search(f"({'|'.join(extra_opts)})", line):
            data[match.group(0)] = to_type(get_numbers(line), float)

    fix_data_types(coords, {'index': int,
                            'u': float,
                            'v': float,
                            'w': float,
                            'mass': float})
    data['coords'] = coords
    return data


def parse_hug_file(hug_file: TextIO) -> Dict[str, List[float]]:
    """ Parse castep .hug file """

    cols = ('compression', 'temperature', 'pressure', 'energy')
    data = defaultdict(list)
    for line in hug_file:
        if match := re.search(labelled_floats(cols), line):
            stack_dict(data, match.groupdict())

    fix_data_types(data, {dt: float for dt in cols})
    return data


def parse_bands_file(bands_file: TextIO) -> Dict[str, Any]:
    """ Parse castep .bonds file """

    bands_info = defaultdict(list)
    qdata = {}
    for line in bands_file:
        if block := get_block(line, bands_file, "BEGIN header", "END header"):

            data = parse_regular_header(block)
            bands_info.update(data)

        elif line.startswith("K-point"):
            if qdata:
                fix_data_types(qdata, {'qpt': float,
                                       'weight': float,
                                       'spin_comp': int,
                                       'band': float
                                       })
                bands_info['bands'].append(qdata)
            _, _, *qpt, weight = line.split()
            qdata = {'qpt': qpt, 'weight': weight, 'spin_comp': None, 'band': []}

        elif line.startswith("Spin component"):
            qdata['spin_comp'] = line.split()[2]
            if qdata['spin_comp'] != "1":
                qdata['band_up'] = qdata.pop('band')

        elif re.match(rf"^\s*{REs.FNUMBER_RE}$", line.strip()):
            if qdata['spin_comp'] != "1":
                qdata['band_dn'].append(line)
            else:
                qdata['band'].append(line)

    if qdata:
        fix_data_types(qdata, {'qpt': float,
                               'weight': float,
                               'spin_comp': int,
                               'band': float,
                               'band_up': float,
                               'band_dn': float
                               })
        bands_info['bands'].append(qdata)

    return bands_info


def parse_phonon_dos_file(phonon_dos_file: TextIO) -> Dict[str, Any]:
    """ Parse castep .phonon_dos file """
    # pylint: disable=too-many-branches,redefined-outer-name
    logger = log_factory(phonon_dos_file)
    phonon_dos_info = defaultdict(list)

    for line in phonon_dos_file:
        if block := get_block(line, phonon_dos_file, "BEGIN header", "END header"):
            data = parse_regular_header(block)
            phonon_dos_info.update(data)

        elif block := get_block(line, phonon_dos_file, "BEGIN GRADIENTS", "END GRADIENTS"):

            logger("Found gradient block")
            qdata = defaultdict(list)

            def fix(qdat):
                fix_data_types(qdat, {'qpt': float,
                                      'pth': float,
                                      'n': int,
                                      'f': float,
                                      'Grad_qf': float})

            for line in block:
                if match := REs.PHONON_PHONON_RE.match(line):
                    if qdata:
                        fix(qdata)
                        phonon_dos_info['gradients'].append(qdata)
                    qdata = defaultdict(list)

                    for key, val in match.groupdict().items():
                        qdata[key] = val.split()

                elif match := REs.PROCESS_PHONON_PHONON_RE.match(line):
                    stack_dict(qdata, match.groupdict())

            if qdata:
                fix(qdata)
                phonon_dos_info['gradients'].append(qdata)

        elif block := get_block(line, phonon_dos_file, "BEGIN DOS", "END DOS", out_fmt=list):

            logger("Found DOS block")

            dos = defaultdict(list)
            for line in block[1:-2]:
                match = re.match(labelled_floats(('freq', 'g', 'si')), line)
                stack_dict(dos, match.groupdict())

            if dos:
                fix_data_types(dos, {'freq': float,
                                     'g': float,
                                     'si': float})
                phonon_dos_info['dos'].append(dos)

    return phonon_dos_info


def parse_efield_file(efield_file: TextIO) -> Dict[str, Union[float, int]]:
    """ Parse castep .efield file """
    # pylint: disable=too-many-branches,redefined-outer-name

    efield_info = defaultdict(list)
    logger = log_factory(efield_file)

    for line in efield_file:
        if block := get_block(line, efield_file, "BEGIN header", "END header"):
            data = parse_regular_header(block, ('Oscillator Q',))
            efield_info.update(data)

        elif block := get_block(line, efield_file, "BEGIN Oscillator Strengths",
                                "END Oscillator Strengths",
                                out_fmt=list):

            logger("Found Oscillator Strengths")

            osc = defaultdict(list)
            for line in block[1:-2]:
                match = re.match(rf"\s*(?P<freq>{REs.INTNUMBER_RE})" +
                                 labelled_floats([*(f'S{d}' for d in SND_D)]), line)
                stack_dict(osc, match.groupdict())

            if osc:
                fix_data_types(osc, {'freq': float,
                                     **{f'S{d}': float for d in SND_D}})
                efield_info['oscillator_strengths'].append(osc)

        elif block := get_block(line, efield_file, "BEGIN permittivity", "END permittivity",
                                out_fmt=list):

            logger("Found permittivity")

            perm = defaultdict(list)
            for line in block[1:-2]:
                match = re.match(labelled_floats(['freq',
                                                  *(f'e_r_{d}' for d in SND_D)]), line)
                stack_dict(perm, match.groupdict())

            if perm:
                fix_data_types(perm, {'freq': float,
                                      **{f'e_r_{d}': float for d in SND_D}})
                efield_info['permittivity'].append(perm)

    return efield_info


def parse_xrd_sf_file(xrd_sf_file: TextIO) -> Dict[str, Union[int, float]]:
    """ Parse castep .xrd_sf file """

    # Get headers from first line
    headers = xrd_sf_file.readline().split()[3:]
    # Turn Re(x) into x_re
    headers = [(head[3:-1]+"_"+head[0:2]).lower() for head in headers]

    xrd_re = re.compile(rf"(?P<qvec>(?:\s*{REs.INTNUMBER_RE}){{3}})" +
                        labelled_floats(headers))

    xrd = defaultdict(list)
    for line in xrd_sf_file:
        match = xrd_re.match(line)
        stack_dict(xrd, match.groupdict())

    if xrd:
        xrd["qvec"] = [[*map(int, qvec)] for x in xrd["qvec"] if (qvec := x.split())]
        fix_data_types(xrd, {head: float for head in headers})

    return xrd


def parse_elastic_file(elastic_file: TextIO) -> Dict[str, List[List[float]]]:
    """ Parse castep .elastic files """
    accum = defaultdict(list)

    for line in elastic_file:
        if block := get_block(line, elastic_file,
                              "^BEGIN Elastic Constants",
                              "^END Elastic Constants"):

            accum["elastic_constants"] = [to_type(numbers, float)
                                          for blk_line in block
                                          if (numbers := get_numbers(blk_line))]

        elif block := get_block(line, elastic_file,
                                "^BEGIN Compliance Matrix",
                                "^END Compliance Matrix"):
            next(block)  # Skip Begin line w/units

            accum["compliance_matrix"] = [to_type(numbers, float)
                                          for blk_line in block
                                          if (numbers := get_numbers(blk_line))]

    return accum


def parse_ts_file(ts_file: TextIO) -> Dict[str, Any]:
    """ Parse castep .ts files """

    accum = defaultdict(list)

    for line in ts_file:
        if "TSConfirmation" in line:
            accum["confirmation"] = True

        elif block := get_block(line, ts_file, "(REA|PRO|TST)", r"^\s*$", eof_possible=True):
            curr = defaultdict(list)
            match = re.match(r"\s*(?P<type>REA|PRO|TST)\s*\d+\s*" +
                             labelled_floats(('reaction_coordinate',)), line)
            key = TS_TYPES[match["type"]]
            curr["reaction_coordinate"] = to_type(match["reaction_coordinate"], float)

            for blk_line in block:
                if match := ATDATTAG.search(blk_line):
                    ion = atreg_to_index(match)
                    if ion not in curr:
                        curr[ion] = {}
                    curr[ion][match.group('tag')] = to_type([*(match.group(d)
                                                               for d in FST_D)], float)
                    add_aliases(curr[ion], TAG_ALIASES)

                elif match := TAG_RE.search(blk_line):
                    curr[match.group('tag')].append([*to_type(get_numbers(blk_line), float)])

            add_aliases(curr, TAG_ALIASES)
            accum[key].append(curr)

    return accum


def parse_kpt_info(inp: TextIO, prop: Union[str, Tuple[str]]) -> Dict[str, List[Union[int, float]]]:
    """ Parse standard form of kpt related .*_fmt files """

    # Skip header
    while "END header" not in inp.readline():
        pass

    qdata = defaultdict(list)
    for line in inp:
        if not line.strip():
            continue
        if isinstance(prop, str):
            *qpt, val = line.split()
            qpt = to_type(qpt, int)
            val = to_type(val, float)
            stack_dict(qdata, {'q': qpt, prop: val})
        elif isinstance(prop, tuple):
            words = line.split()
            qpt = to_type(words[0:3], int)
            val = to_type(words[3:], float)
            stack_dict(qdata, {'q': qpt, **dict(zip(prop, val))})

    return qdata


def parse_elf_fmt_file(elf_file: TextIO) -> Dict[str, List[Union[int, float]]]:
    """ Parse castep .elf_fmt files """
    return parse_kpt_info(elf_file, ('chi_alpha', 'chi_beta'))


def parse_chdiff_fmt_file(chdiff_file: TextIO) -> Dict[str, List[Union[int, float]]]:
    """ Parse castep .chdiff_fmt files """
    return parse_kpt_info(chdiff_file, 'chdiff')


def parse_pot_fmt_file(pot_file: TextIO) -> Dict[str, List[Union[int, float]]]:
    """ Parse castep .pot_fmt files """
    return parse_kpt_info(pot_file, 'pot')


def parse_den_fmt_file(den_file: TextIO) -> Dict[str, List[Union[int, float]]]:
    """ Parse castep .den_fmt files """
    return parse_kpt_info(den_file, 'density')



"""
Parse castep .cell and .param files
"""
from typing import Dict, TextIO, List, Union
import re
from .utility import log_factory
from .castep_res import get_block, DEVEL_CODE_BLOCK_RE, DEVEL_CODE_VAL_RE


def _parse_devel_code_block(in_block: str) -> Dict[str, str]:
    """ Parse devel_code block to dict """

    matches = re.finditer(DEVEL_CODE_BLOCK_RE, in_block, re.IGNORECASE | re.MULTILINE)

    devel_code_parsed = {}

    # Get groups
    for blk in matches:
        block_title = blk.group(1)
        block = {}
        for par in re.finditer(DEVEL_CODE_VAL_RE, blk.group(0),
                               re.IGNORECASE | re.MULTILINE):

            key, val = re.split('[:=]', par.group(0))
            block[key] = val

        devel_code_parsed[block_title] = block

        # Remove matched to get remainder
        in_block = in_block.replace(blk.group(0), '')

    # Catch remainder
    for par in re.finditer(DEVEL_CODE_VAL_RE, in_block):
        key, val = re.split('[:=]', par.group(0))
        devel_code_parsed[key] = val

    return devel_code_parsed


def parse_cell_param_file(cell_param_file: TextIO) -> List[Dict[str, Union[str, Dict[str, str]]]]:
    """ Parse .cell/.param files into dict ready to JSONise """

    logger = log_factory(cell_param_file)
    curr = {}

    for line in cell_param_file:
        # Strip comments
        line = re.split('[#!]', line)[0]

        if block := get_block(line, cell_param_file,
                              re.compile('^%block ', re.IGNORECASE),
                              re.compile('^%endblock', re.IGNORECASE),
                              out_fmt=str):

            block_title = block.splitlines()[0].split()[1].lower()

            logger("Found block %s", block_title)

            if block_title == 'devel_code':
                curr['devel_code'] = _parse_devel_code_block(block)
            else:
                block = block.splitlines()
                curr[block_title] = [*map(lambda x: x.strip().split(), block[1:-1])]
        elif line := line.strip():

            key, val = re.split(r'\s*[ :=]+\s*', line, maxsplit=1)

            key = key.lower()
            logger("Found param %s", key)

            curr[key] = val
    return [curr]


parse_cell_file = parse_cell_param_file
parse_param_file = parse_cell_param_file



"""
Module for constants used in castep_outputs
"""

from typing import Literal, Tuple

SHELLS = ('s', 'p', 'd', 'f')
FST_D = ('x', 'y', 'z')
SND_D = ('xx', 'yy', 'zz', 'yz', 'zx', 'xy')
MINIMISERS = ('bfgs', 'lbfgs', 'fire', 'tpsd', 'dmd')
PAIR_POTS = ('LJ', 'BUCK', 'COUL', 'DZ', 'POL', 'BB', 'SHO',
             'SW', 'MORS', 'POLM', 'LJ_S', 'PES', 'BU_S', 'TIP4', 'QUIP')

TAG_ALIASES = {'E': 'energy',
               'T': 'temperature',
               'P': 'pressure',
               'h': 'lattice_vectors',
               'hv': 'lattice_velocity',
               'R': 'position',
               'V': 'velocity',
               'F': 'force'}

TS_TYPES = {"REA": "reagent",
            "PRO": "product",
            "TST": "test"}


CASTEP_OUTPUT_NAMES = (
    "castep",
    "param",
    "cell",
    "geom",
    "md",
    "bands",
    "hug",
    "phonon_dos",
    "efield",
    "xrd_sf",
    "elf_fmt",
    "chdiff_fmt",
    "pot_fmt",
    "den_fmt",
    "elastic",
    "ts"
)
CASTEP_FILE_FORMATS = tuple(f".{typ}" for typ in CASTEP_OUTPUT_NAMES)

OutFormats = Literal["json", "pprint", "print", "ruamel", "yaml"]
AtomIndex = Tuple[str, float]
ThreeVector = Tuple[float, float, float]
ThreeByThreeMatrix = Tuple[ThreeVector, ThreeVector, ThreeVector]



# pylint: disable=too-many-lines, too-many-branches, too-many-statements
"""
Extract results from .castep file for comparison and use
by testcode.pl.

Port of extract_results.pl
"""

from collections import defaultdict
from typing import TextIO, List, Dict, Any, Union, Sequence, Tuple, Optional
import io
import re
import itertools

from . import castep_res as REs
from .castep_res import labelled_floats, get_numbers, get_block, gen_table_re

from .constants import SHELLS, ThreeVector, ThreeByThreeMatrix, AtomIndex

from .utility import (fix_data_types, add_aliases, to_type,
                      stack_dict, normalise_string,
                      atreg_to_index, log_factory)
from .extra_files_parser import (parse_bands_file,
                                 parse_hug_file,
                                 parse_phonon_dos_file,
                                 parse_efield_file,
                                 parse_xrd_sf_file,
                                 parse_elf_fmt_file,
                                 parse_chdiff_fmt_file,
                                 parse_pot_fmt_file,
                                 parse_den_fmt_file,
                                 parse_elastic_file)


def parse_castep_file(castep_file: TextIO) -> List[Dict[str, Any]]:
    """ Parse castep file into lists of dicts ready to JSONise """
    # pylint: disable=redefined-outer-name

    runs = []
    curr_run = defaultdict(list)

    logger = log_factory(castep_file)

    for line in castep_file:
        # Build Info
        if block := get_block(line, castep_file,
                              r"^\s*Compiled for",
                              REs.EMPTY, out_fmt=list):

            if curr_run:
                runs.append(curr_run)

            logger("Found build info")
            curr_run = defaultdict(list)
            curr_run["build_info"] = _process_buildinfo(block)

        elif re.search(r"Run started", line):
            curr_run["time_started"] = normalise_string(line.split(":", 1)[1])

            logger("Found run %s", len(runs) + 1)

        # Finalisation
        elif block := get_block(line, castep_file, "Initialisation time", REs.EMPTY):

            logger("Found finalisation")

            curr_run.update(_process_finalisation(block))

        elif line.startswith("Overall parallel efficiency rating"):

            logger("Found parallel efficiency")

            curr_run["parallel_efficiency"] = float(get_numbers(line)[0])

        # Title
        elif re.match(gen_table_re("Title", r"\*+"), line):

            logger("Found title")

            curr_run["title"] = next(castep_file).strip()

        # Memory estimate
        elif block := get_block(line, castep_file,
                                gen_table_re(r"MEMORY AND SCRATCH[\w\s]+", "[+-]+"),
                                gen_table_re("", "[+-]+")):

            logger("Found memory estimate")

            curr_run["memory_estimate"].append(_process_memory_est(block))

        # Parameters
        elif block := get_block(line, castep_file,
                                gen_table_re("[^*]+ Parameters", r"\*+"),
                                gen_table_re("", r"\*+")):

            logger("Found options")

            curr_run["options"] = _process_params(block)

        # Pseudo-atomic energy
        elif block := get_block(line, castep_file, REs.PS_SHELL_RE, REs.EMPTY, cnt=2):

            logger("Found pseudo-atomic energy")

            key, val = _process_ps_energy(block)

            if "species_properties" not in curr_run:
                curr_run["species_properties"] = defaultdict(dict)

            curr_run["species_properties"][key]["pseudo_atomic_energy"] = val

        # Mass
        elif block := get_block(line, castep_file, r"Mass of species in AMU", REs.EMPTY):

            logger("Found mass")

            if "species_properties" not in curr_run:
                curr_run["species_properties"] = defaultdict(dict)

            for key, val in _process_spec_prop(block):
                curr_run["species_properties"][key]["mass"] = float(val)

        # Electric Quadrupole Moment
        elif block := get_block(line, castep_file, r"Electric Quadrupole Moment", r"^\s+[^ A-Z]"):
            logger("Found electric quadrupole moment")

            if "species_properties" not in curr_run:
                curr_run["species_properties"] = defaultdict(dict)

            for key, val, *_ in _process_spec_prop(block):
                curr_run["species_properties"][key]["electric_quadrupole_moment"] = float(val)

        # Pseudopots
        elif block := get_block(line, castep_file, r"Files used for pseudopotentials", REs.EMPTY):

            logger("Found pseudopotentials")

            if "species_properties" not in curr_run:
                curr_run["species_properties"] = defaultdict(dict)

            for key, val in _process_spec_prop(block):
                if "|" in val:
                    val = _process_pspot_string(val)

                curr_run["species_properties"][key]["pseudopot"] = val

        elif block := get_block(line, castep_file,
                                gen_table_re("Pseudopotential Report[^|]+", r"\|"),
                                gen_table_re("", "=+")):

            logger("Found pseudopotential report")

            curr_run["pspot_detail"].append(_process_pspot_report(block))

        elif match := re.match(r"\s*(?P<type>AE|PS) eigenvalue nl (?P<nl>\d+) =" +
                               labelled_floats(('eigenvalue',)), line):

            logger("Found PSPot debug for %s at %s", match['type'], match['nl'])

            match = match.groupdict()
            fix_data_types(match, {'nl': int, 'eigenvalue': float})

            curr_run["pspot_debug"].append(match)

        # Pair Params
        elif block := get_block(line, castep_file,
                                gen_table_re("PairParams", r"\*+", pre=r"\w*"),
                                REs.EMPTY):

            logger("Found pair params")

            curr_run["pair_params"].append(_process_pair_params(block))

        # DFTD
        elif block := get_block(line, castep_file,
                                "DFT-D parameters",
                                gen_table_re("", "x+"), cnt=2):

            logger("Found DFTD block")

            curr_run["dftd"] = _process_dftd(block)

        # SCF
        elif block := get_block(line, castep_file, "SCF loop", "^-+ <-- SCF", cnt=2):

            logger("Found SCF")

            curr_run["scf"].append(_process_scf(block))

        # Line min
        elif block := get_block(line, castep_file,
                                gen_table_re("WAVEFUNCTION LINE MINIMISATION", "[+-]+",
                                             post="<- line"),
                                gen_table_re("", "[+-]+", post="<- line"), cnt=2):

            logger("Found wvfn line min")

            curr_run["wvfn_line_min"].append(_process_wvfn_line_min(block))

        # Occupancy
        elif block := get_block(line, castep_file,
                                gen_table_re("Occupancy", r"\|",
                                             post="<- occ", whole_line=False),
                                r"Have a nice day\."):

            logger("Found occupancies")

            curr_run["occupancies"].append(_process_occupancies(block))

        # Energies
        elif line.startswith("Final energy, E") or line.startswith("Final energy"):

            logger("Found energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["final_energy"].append(to_type(get_numbers(line)[-1], float))

        elif "Total energy corrected for finite basis set" in line:

            logger("Found energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["final_basis_set_corrected"].append(
                to_type(get_numbers(line)[-1], float))

        elif "0K energy (E-0.5TS)" in line:

            logger("Found estimated 0K energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["est_0K"].append(to_type(get_numbers(line)[-1], float))

        elif line.startswith("(SEDC) Total Energy"):

            logger("Found SEDC energy correction")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["sedc_correction"].append(to_type(get_numbers(line)[-1], float))

        elif line.startswith("Dispersion corrected final energy"):

            logger("Found SEDC final energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["disperson_corrected"].append(
                to_type(get_numbers(line)[-1], float))

        # Free energies
        elif re.match(rf"Final free energy \(E-TS\) += +({REs.EXPNUMBER_RE})", line):

            logger("Found free energy (E-TS)")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["free_energy"].append(to_type(get_numbers(line)[-1], float))

        # Solvation energy
        elif line.startswith(" Free energy of solvation"):

            logger("Found solvation energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["solvation"].append(*to_type(get_numbers(line), float))

        # Spin densities
        elif match := re.search(rf"Integrated \|?Spin Density\|?\s+=\s+({REs.EXPNUMBER_RE})", line):

            logger("Found spin")

            if "|" in line:
                curr_run["modspin"].append(to_type(match.group(1), float))
            else:
                curr_run["spin"].append(to_type(match.group(1), float))

        # Initial cell
        elif block := get_block(line, castep_file, gen_table_re("Unit Cell"), REs.EMPTY, cnt=3):

            logger("Found cell")

            curr_run["initial_cell"] = _process_unit_cell(block)

        # Cell Symmetry and contstraints
        elif block := get_block(line, castep_file,
                                gen_table_re("Symmetry and Constraints"),
                                "Cell constraints are"):

            logger("Found symmetries")

            curr_run["symmetries"], curr_run["constraints"] = _process_symmetry(block)

        # TSS (must be ahead of initial pos)
        elif block := get_block(line, castep_file,
                                gen_table_re("(Reactant|Product)", "x"),
                                gen_table_re("", "x+"), cnt=2):

            mode = "reactant" if "Reactant" in line else "product"

            logger("Found %s initial states", mode)

            curr_run[mode] = _process_atreg_block(block)

        # Initial pos
        elif block := get_block(line, castep_file,
                                "Fractional coordinates of atoms",
                                gen_table_re("", "x+")):

            logger("Found initial positions")

            curr_run["initial_positions"] = _process_atreg_block(block)

        elif "Supercell generated" in line:
            accum = iter(get_numbers(line))
            curr_run["supercell"] = tuple(to_type([next(accum) for _ in range(3)], float)
                                          for _ in range(3))

        # Initial vel
        elif block := get_block(line, castep_file,
                                "User Supplied Ionic Velocities",
                                gen_table_re("", "x+")):

            logger("Found initial velocities")

            curr_run["initial_velocities"] = _process_atreg_block(block)

        # Initial spins
        elif block := get_block(line, castep_file,
                                "Initial magnetic",
                                gen_table_re("", "x+")):

            logger("Found initial spins")

            curr_run["initial_spins"] = _process_initial_spins(block)

        # Target Stress
        elif block := get_block(line, castep_file, "External pressure/stress", REs.EMPTY):

            logger("Found target stress")

            curr_run["target_stress"].extend(number
                                             for line in block
                                             for number in to_type(get_numbers(line), float))

        # Finite basis correction parameter
        elif match := re.search(rf"finite basis dEtot\/dlog\(Ecut\) = +({REs.FNUMBER_RE})", line):

            logger("Found dE/dlog(E)")
            curr_run["dedlne"].append(to_type(match.group(1), float))

        # K-Points
        elif block := get_block(line, castep_file, "k-Points For BZ Sampling", REs.EMPTY):

            logger("Found k-points")

            curr_run["k-points"] = _process_kpoint_blocks(block, True)

        elif block := get_block(line, castep_file,
                                gen_table_re("Number +Fractional coordinates +Weight", r"\+"),
                                gen_table_re("", r"\++")):

            logger("Found k-points list")

            curr_run["k-points"] = _process_kpoint_blocks(block, False)

        # Forces blocks
        elif block := get_block(line, castep_file, REs.FORCES_BLOCK_RE, r"^\s*\*+$"):
            if "forces" not in curr_run:
                curr_run["forces"] = defaultdict(list)

            key, val = _process_forces(block)

            logger("Found %s forces", key)

            curr_run["forces"][key].append(val)

        # Stress tensor block
        elif block := get_block(line, castep_file, REs.STRESSES_BLOCK_RE, r"^\s*\*+$"):
            if "stresses" not in curr_run:
                curr_run["stresses"] = defaultdict(list)

            key, val = _process_stresses(block)

            logger("Found %s stress", key)

            curr_run["stresses"][key].append(val)

        # Phonon block
        elif match := REs.PHONON_RE.match(line):

            logger("Found phonon")

            qdata = defaultdict(list)

            qdata["qpt"] = match.group("qpt").split()

            logger("Reading qpt %s", qdata['qpt'], level="debug")

            for line in castep_file:
                if match := REs.PHONON_RE.match(line):
                    if qdata["qpt"] and qdata["qpt"] not in (phonon["qpt"]
                                                             for phonon in curr_run["phonons"]):
                        curr_run["phonons"].append(_process_qdata(qdata))
                    qdata = defaultdict(list)
                    qdata["qpt"] = match.group("qpt").split()

                    logger("Reading qpt %s", qdata['qpt'], level="debug")

                elif (re.match(r"\s+\+\s+Effective cut-off =", line) or
                      re.match(rf"\s+\+\s+q->0 along \((\s*{REs.FNUMBER_RE}){{3}}\)\s+\+", line) or
                      re.match(r"\s+\+ -+ \+", line)):
                    continue
                elif match := REs.PROCESS_PHONON_RE.match(line):

                    # ==By mode
                    # qdata["modes"].append(match.groupdict())
                    # ==By prop
                    stack_dict(qdata, match.groupdict())

                elif re.match(r"\s+\+\s+.*\+", line):
                    continue
                else:
                    break

            else:
                raise IOError(f"Unexpected end of file in {castep_file.name}")

            if qdata["qpt"] and qdata["qpt"] not in (phonon["qpt"]
                                                     for phonon in curr_run["phonons"]):
                curr_run["phonons"].append(_process_qdata(qdata))

            logger("Found %d phonon samples", len(curr_run['phonons']))

        # Phonon Symmetry
        elif block := get_block(line, castep_file,
                                "Phonon Symmetry Analysis", REs.EMPTY):

            logger("Found phonon symmetry analysis")

            val = _process_phonon_sym_analysis(block)
            curr_run["phonon_symmetry_analysis"].append(val)

            # Solvation
        elif block := get_block(line, castep_file,
                                gen_table_re("AUTOSOLVATION CALCULATION RESULTS", r"\*+"),
                                r"^\s*\*+\s*$"):

            logger("Found autosolvation")

            curr_run["autosolvation"] = _process_autosolvation(block)

        # Dynamical Matrix
        elif block := get_block(line, castep_file,
                                gen_table_re("Dynamical matrix"),
                                gen_table_re("", "-+")):

            logger("Found dynamical matrix")

            val = _process_dynamical_matrix(block)
            curr_run["dynamical_matrix"] = val

        # Raman tensors
        elif block := get_block(line, castep_file,
                                gen_table_re("Raman Susceptibility Tensors[^+]*", r"\+"),
                                REs.EMPTY):

            logger("Found Raman")

            curr_run["raman"].append(_process_raman(block))

        # Born charges
        elif block := get_block(line, castep_file,
                                gen_table_re("Born Effective Charges"),
                                gen_table_re("", "=+")):

            logger("Found Born")

            curr_run["born"].append(_process_born(block))

        # Permittivity and NLO Susceptibility
        elif block := get_block(line, castep_file,
                                r"^\s+Optical Permittivity", r"^ =+$"):

            logger("Found optical permittivity")

            val = _process_3_6_matrix(block, True)
            curr_run["optical_permittivity"] = val[0]
            if val[1]:
                curr_run["dc_permittivity"] = val[1]

        # Polarisability
        elif block := get_block(line, castep_file, r"^\s+Polarisabilit(y|ies)", r"^ =+$"):

            logger("Found polarisability")

            val = _process_3_6_matrix(block, True)
            curr_run["optical_polarisability"] = val[0]
            if val[1]:
                curr_run["static_polarisability"] = val[1]

        # Non-linear
        elif block := get_block(line, castep_file,
                                r"^\s+Nonlinear Optical Susceptibility", r"^ =+$"):

            logger("Found NLO")

            curr_run["nlo"], _ = _process_3_6_matrix(block, False)

        # Thermodynamics
        elif block := get_block(line, castep_file,
                                gen_table_re("Thermodynamics"),
                                gen_table_re("", "-+"), cnt=3):

            logger("Found thermodynamics")

            accum = _process_thermodynamics(block)
            curr_run["thermodynamics"] = accum

        # Mulliken Population Analysis
        elif block := get_block(line, castep_file,
                                gen_table_re(r"Atomic Populations \(Mulliken\)"),
                                gen_table_re("", "=+"), cnt=2):

            logger("Found Mulliken")

            curr_run["mulliken_popn"] = _process_mulliken(block)

        # Orbital populations
        elif block := get_block(line, castep_file,
                                gen_table_re("Orbital Populations"),
                                "The total projected population"):

            logger("Found Orbital populations")

            curr_run["orbital_popn"] = _process_orbital_populations(block)

        # Bond analysis
        elif block := get_block(line, castep_file,
                                r"Bond\s+Population\s+Length",
                                gen_table_re("", "=+"), cnt=2):

            logger("Found bond info")

            curr_run["bonds"] = _process_bond_analysis(block)

        # Hirshfeld Population Analysis
        elif block := get_block(line, castep_file,
                                gen_table_re("Hirshfeld Analysis"),
                                gen_table_re("", "=+"), cnt=2):

            logger("Found Hirshfeld")

            curr_run["hirshfeld"] = _process_hirshfeld(block)

        # ELF
        elif block := get_block(line, castep_file,
                                gen_table_re("ELF grid sample"),
                                gen_table_re("", "-+"), cnt=2):

            logger("Found ELF")

            curr_run["elf"] = _process_elf(block)

        # MD Block
        elif block := get_block(line, castep_file,
                                gen_table_re("MD Data:", "x"),
                                gen_table_re("", "x+")):

            logger("Found MD Block (step %d)", len(curr_run['md'])+1)

            curr_run["md"].append(_process_md(block))

        # GeomOpt
        elif block := get_block(line, castep_file, "Final Configuration",
                                gen_table_re("", "x+"), cnt=2):

            if "geom_opt" not in curr_run:
                curr_run["geom_opt"] = defaultdict(list)

            logger("Found final geom configuration")

            curr_run["geom_opt"]["final_configuration"] = _process_atreg_block(block)

        elif match := re.search(f"(?P<minim>{REs.MINIMISERS_RE}):"
                                r" finished iteration\s*\d+\s*with enthalpy", line):

            if "geom_opt" not in curr_run:
                curr_run["geom_opt"] = defaultdict(list)

            minim = match["minim"]

            logger("Found %s energy", minim)

            curr_run["geom_opt"]["enthalpy"].append(to_type(get_numbers(line)[-1], float))

        elif match := re.match(rf"^\s*(?:{REs.MINIMISERS_RE}):"
                               r"(?P<key>[^=]+)=\s*"
                               f"(?P<value>{REs.EXPNUMBER_RE}).*",
                               line, re.IGNORECASE):

            if "geom_opt" not in curr_run:
                curr_run["geom_opt"] = defaultdict(list)

            key, val = normalise_string(match["key"]).lower(), to_type(match["value"], float)

            logger("Found geomopt %s", key)

            curr_run["geom_opt"][key] = val

        elif block := get_block(line, castep_file,
                                f"<--( min)? {REs.MINIMISERS_RE}$", r"\+(?:-+\+){4,5}", cnt=2):

            typ = re.search(REs.MINIMISERS_RE, line).group(0)

            if "geom_opt" not in curr_run:
                curr_run["geom_opt"] = defaultdict(list)

            logger("Found %s geom_block", typ)

            curr_run["geom_opt"]["minimisation"].append(_process_geom_table(block))

        # TDDFT
        elif block := get_block(line, castep_file,
                                gen_table_re("TDDFT excitation energies", r"\+", post="TDDFT"),
                                gen_table_re("=+", r"\+", post="TDDFT"), cnt=2):

            logger("Found TDDFT excitations")

            curr_run["tddft"] = _process_tddft(block)

        # Band structure
        elif block := get_block(line, castep_file,
                                gen_table_re("(B A N D|Band Structure Calculation)[^+]+", r"\+"),
                                gen_table_re("", "=+")):

            logger("Found band-structure")

            curr_run["bs"] = _process_band_structure(block)

            # Molecular Dipole
        elif block := get_block(line, castep_file,
                                gen_table_re("D I P O L E   O F   M O L E C U L E"
                                             "   I N   S U P E R C E L L",
                                             r"\+"),
                                gen_table_re("", "=+")):

            logger("Found molecular dipole")

            curr_run["molecular_dipole"] = _process_dipole(block)

        # Chemical shielding
        elif block := get_block(line, castep_file,
                                gen_table_re("Chemical Shielding Tensor", r"\|"),
                                gen_table_re("", "=+")):

            logger("Found Chemical Shielding Tensor")

            val = _parse_magres_block(0, block)
            curr_run["magres"].append(val)

        elif block := get_block(line, castep_file,
                                gen_table_re("Chemical Shielding and "
                                             "Electric Field Gradient Tensors", r"\|"),
                                gen_table_re("", "=+")):

            logger("Found Chemical Shielding + EField Tensor")

            val = _parse_magres_block(1, block)
            curr_run["magres"].append(val)

        elif block := get_block(line, castep_file,
                                gen_table_re("Electric Field Gradient Tensor", r"\|"),
                                gen_table_re("", "=+")):

            logger("Found EField Tensor")

            val = _parse_magres_block(2, block)
            curr_run["magres"].append(val)

        elif block := get_block(line, castep_file,
                                gen_table_re("(?:I|Ani)sotropic J-coupling", r"\|"),
                                gen_table_re("", "=+")):

            logger("Found J-coupling")

            val = _parse_magres_block(3, block)
            curr_run["magres"].append(val)

        elif block := get_block(line, castep_file,
                                gen_table_re("Hyperfine Tensor", r"\|"),
                                gen_table_re("", "=+")):

            logger("Found Hyperfine tensor")

            val = _parse_magres_block(4, block)
            curr_run["magres"].append(val)

        # Elastic
        elif block := get_block(line, castep_file,
                                gen_table_re(r"Elastic Constants Tensor \(GPa\)"),
                                gen_table_re("", "=+")):

            logger("Found elastic constants tensor")

            if "elastic" not in curr_run:
                curr_run["elastic"] = {}
            val, _ = _process_3_6_matrix(block, False)
            curr_run["elastic"]["elastic_constants"] = val

        elif block := get_block(line, castep_file,
                                gen_table_re(r"Compliance Matrix \(GPa\^-1\)"),
                                gen_table_re("", "=+")):

            logger("Found compliance matrix")

            if "elastic" not in curr_run:
                curr_run["elastic"] = {}
            val, _ = _process_3_6_matrix(block, False)
            curr_run["elastic"]["compliance_matrix"] = val

        elif block := get_block(line, castep_file, "Contribution ::", REs.EMPTY):
            typ = re.match("(?P<type>.* Contribution)", line).group("type")
            next(block)

            logger("Found elastic %s contribution", typ)

            if "elastic" not in curr_run:
                curr_run["elastic"] = {}

            val, _ = _process_3_6_matrix(block, False)
            curr_run["elastic"][typ] = val

        elif block := get_block(line, castep_file,
                                gen_table_re("Elastic Properties"),
                                gen_table_re("", "=+")):

            logger("Found elastic properties")

            if "elastic" not in curr_run:
                curr_run["elastic"] = {}

            curr_run["elastic"].update(_process_elastic_properties(block))

        # --- Extra blocks for testing

        # Hugoniot data
        elif block := get_block(line, castep_file, "BEGIN hug", "END hug"):

            logger("Found hug block")

            val = parse_hug_file(block)
            curr_run["hug"].append(val)

        # Bands block (spectral data)
        elif block := get_block(line, castep_file, "BEGIN bands", "END bands"):

            logger("Found bands block")

            val = parse_bands_file(block)
            curr_run["bands"].append(val["bands"])

        elif block := get_block(line, castep_file, "BEGIN phonon_dos", "END phonon_dos"):

            logger("Found phonon_dos block")

            val = parse_phonon_dos_file(block)
            curr_run["phonon_dos"] = val["dos"]
            curr_run["gradients"] = val["gradients"]

        # E-Field
        elif block := get_block(line, castep_file, "BEGIN efield", "END efield"):

            logger("Found efield block")

            val = parse_efield_file(block)
            curr_run["oscillator_strengths"] = val["oscillator_strengths"]
            curr_run["permittivity"] = val["permittivity"]

        # Elastic
        elif block := get_block(line, castep_file, "<BEGIN elastic>", "<END elastic>"):

            logger("Found elastic block")

            val = parse_elastic_file(block)
            curr_run["oscillator_strengths"] = val["elastic_constants"]
            curr_run["permittivity"] = val["compliance_matrix"]

        # XRD Structure Factor
        elif block := get_block(line, castep_file, "BEGIN xrd_sf", "END xrd_sf",
                                out_fmt=list):

            logger("Found xrdsf")

            block = "\n".join(block[1:-1])  # Strip begin/end tags lazily
            block = io.StringIO(block)
            val = parse_xrd_sf_file(block)

            curr_run["xrd_sf"] = val

        # ELF FMT
        elif block := get_block(line, castep_file, "BEGIN elf_fmt", "END elf_fmt",
                                out_fmt=list):

            logger("Found ELF fmt")

            block = "\n".join(block[1:-1])  # Strip begin/end tags lazily
            block = io.StringIO(block)
            val = parse_elf_fmt_file(block)
            if "kpt-data" not in curr_run:
                curr_run["kpt-data"] = val
            else:
                curr_run["kpt-data"].update(val)

        # CHDIFF FMT
        elif block := get_block(line, castep_file, "BEGIN chdiff_fmt", "END chdiff_fmt",
                                out_fmt=list):

            logger("Found CHDIFF fmt")

            block = "\n".join(block[1:-1])  # Strip begin/end tags lazily
            block = io.StringIO(block)
            val = parse_chdiff_fmt_file(block)
            if "kpt-data" not in curr_run:
                curr_run["kpt-data"] = val
            else:
                curr_run["kpt-data"].update(val)

        # POT FMT
        elif block := get_block(line, castep_file, "BEGIN pot_fmt", "END pot_fmt",
                                out_fmt=list):

            logger("Found POT fmt")

            block = "\n".join(block[1:-1])  # Strip begin/end tags lazily
            block = io.StringIO(block)
            val = parse_pot_fmt_file(block)
            if "kpt-data" not in curr_run:
                curr_run["kpt-data"] = val
            else:
                curr_run["kpt-data"].update(val)

        # DEN FMT
        elif block := get_block(line, castep_file, "BEGIN den_fmt", "END den_fmt",
                                out_fmt=list):

            logger("Found DEN fmt")

            block = "\n".join(block[1:-1])  # Strip begin/end tags lazily
            block = io.StringIO(block)
            val = parse_den_fmt_file(block)
            if "kpt-data" not in curr_run:
                curr_run["kpt-data"] = val
            else:
                curr_run["kpt-data"].update(val)

    if curr_run:
        fix_data_types(curr_run, {"energies": float,
                                  "solvation": float})
        runs.append(curr_run)
    return runs


def _process_ps_energy(block: TextIO) -> Tuple[str, float]:
    match = REs.PS_SHELL_RE.search(next(block))
    spec = match["spec"]
    next(block)
    energy = get_numbers(next(block))[1]
    return spec, float(energy)


def _process_tddft(block: TextIO) -> List[Dict[str, Union[str, float]]]:
    tddata = [{"energy": float(match["energy"]),
               "error": float(match["error"]),
               "type": match["type"]}
              for line in block
              if (match := REs.TDDFT_RE.match(line))]
    return tddata


def _process_atreg_block(block: TextIO) -> Dict[AtomIndex, ThreeVector]:
    accum = {atreg_to_index(match): to_type(match.group("x", "y", "z"), float)
             for line in block
             if (match := REs.ATDAT3VEC.search(line))}
    return accum


def _process_spec_prop(block: TextIO) -> List[List[str]]:

    accum = []

    for line in block:
        words = line.split()
        if words and re.match(rf"{REs.SPECIES_RE}\b", words[0]):

            accum.append(words)

    return accum


def _process_md(block: TextIO) -> Dict[str, float]:
    curr_data = {match.group("key").strip(): float(match.group("val"))
                 for line in block
                 if (match := re.search(r"x\s+"
                                        r"(?P<key>[a-zA-Z][A-Za-z ]+):\s*"
                                        rf"(?P<val>{REs.FNUMBER_RE})", line))}

    return {normalise_string(key): val for key, val in curr_data.items()}


def _process_elf(block: TextIO) -> List[float]:
    curr_data = [to_type(match.group(1), float) for line in block
                 if (match := re.match(rf"\s+ELF\s+\d+\s+({REs.FNUMBER_RE})", line))]
    return curr_data


def _process_hirshfeld(block: TextIO) -> Dict[AtomIndex, float]:
    """ Process Hirshfeld block to dict of charges """
    accum = {atreg_to_index(match): float(match["charge"]) for line in block
             if (match := re.match(rf"\s+{REs.ATREG}\s+(?P<charge>{REs.FNUMBER_RE})", line))}
    return accum


def _process_thermodynamics(block: TextIO) -> Dict[str, List[float]]:
    """ Process a thermodynamics block into a dict of lists """
    accum = defaultdict(list)
    for line in block:
        if "Zero-point energy" in line:
            accum["zero-point_energy"] = float(get_numbers(line)[0])

        if match := REs.THERMODYNAMICS_DATA_RE.match(line):
            stack_dict(accum, match.groupdict())
        # elif re.match(r"\s+T\(", line):  # Can make dict/re based on labels
        #     thermo_label = line.split()

    fix_data_types(accum, {key: float for
                           key in ("T", "E", "F", "S", "Cv")})
    return accum


def _process_3_6_matrix(block: TextIO, split: bool) -> Tuple[ThreeByThreeMatrix,
                                                             Optional[ThreeByThreeMatrix]]:
    """ Process a single or pair of 3x3 matrices or 3x6 matrix """
    fst = [to_type(vals, float) for line in block
           if (vals := get_numbers(line)) and len(vals) in (3, 6)]
    if split and len(fst[0]) == 6:
        fst, snd = [tuple(line[0:3]) for line in fst], [tuple(line[3:6]) for line in fst]
    else:
        snd = []

    return fst, snd


def _process_params(block: TextIO) -> Dict[str, str]:
    """ Process a parameters block into a dict of params """

    opt = {}
    curr_opt = {}
    curr_group = ""

    for line in block:
        if match := re.match(r"\s*\*+ ([A-Za-z ]+) Parameters \*+", line):
            if curr_opt:
                opt[curr_group] = curr_opt
            curr_group = normalise_string(match.group(1)).lower()
            curr_opt = {}
        elif len(match := line.split(":")) > 1:
            *key, val = map(normalise_string, match)
            curr_opt[" ".join(key).strip()] = val.strip()

    if curr_opt:
        opt[curr_group] = curr_opt

    return opt


def _process_buildinfo(block: Sequence[str]) -> Dict[str, str]:
    info = {}

    info["summary"] = " ".join(map(normalise_string, block[0:2]))
    for line in block[2:]:
        if ":" in line:
            key, val = map(normalise_string, line.split(":", 1))
            info[key.strip()] = val.strip()
    return info


def _process_unit_cell(block: TextIO) -> Dict[str, Union[ThreeVector, ThreeByThreeMatrix]]:
    cell = defaultdict(list)
    prop = []
    for line in block:
        numbers = get_numbers(line)
        if len(numbers) == 6:
            cell["real_lattice"].append(to_type(numbers[0:3], float))
            cell["recip_lattice"].append(to_type(numbers[3:6], float))
        elif len(numbers) == 2:
            if any(ang in line for ang in ("alpha", "beta", "gamma")):
                cell["lattice_parameters"].append(to_type(numbers[0], float))
                cell["cell_angles"].append(to_type(numbers[1], float))
            else:
                prop.append(to_type(numbers[0], float))

    cell.update({name: val for val, name in zip(prop, ("volume", "density_amu", "density_g"))})

    return cell


def _process_scf(block: TextIO) -> Dict[str, Any]:
    scf = []
    curr = {}
    for line in block:
        if match := REs.SCF_LOOP_RE.match(line):
            if curr:
                scf.append(curr)
            curr = match.groupdict()
            fix_data_types(curr, {"energy": float,
                                  "energy_gain": float,
                                  "fermi_energy": float,
                                  "time": float})

        elif "Density was not mixed" in line:
            curr["density_residual"] = None

        elif "Norm of density" in line:
            curr["density_residual"] = to_type(get_numbers(line)[0], float)

        elif "no. bands" in line:
            curr["no_bands"] = to_type(get_numbers(line)[0], int)

        elif "Kinetic eigenvalue" in line:
            if "kinetic_eigenvalue" not in curr:
                curr["kinetic_eigenvalue"] = []
                curr["eigenvalue"] = []

            curr["kinetic_eigenvalue"] = to_type(get_numbers(line)[1], float)
            eig = []

        elif re.match(r"eigenvalue\s*\d+\s*init=", line):
            labels = ("initial", "final", "change")
            numbers = get_numbers(line)
            eig.append({key: float(val) for val, key in zip(numbers[1:], labels)})

        elif "Checking convergence criteria" in line:
            curr["eigenvalue"].append(eig)
            eig = []

        elif match := re.match(r"[+(]?(?P<key>[()0-9A-Za-z -]+)="
                               rf"\s*{labelled_floats(('val',))} eV\)?", line):
            key, val = normalise_string(match["key"]).lower(), float(match["val"])
            curr[key] = val

    if curr:
        scf.append(curr)

    return scf


def _process_forces(block: TextIO) -> Tuple[str, Dict[AtomIndex, ThreeVector]]:
    ftype = (ft_guess if (ft_guess := REs.FORCES_BLOCK_RE.search(next(block)).group(1))
             else "non-descript")

    ftype = normalise_string(ftype).lower()

    accum = _process_atreg_block(block)

    return ftype, accum


def _process_stresses(block: TextIO) -> Tuple[float]:

    ftype = (ft_guess if (ft_guess := REs.STRESSES_BLOCK_RE.search(next(block)).group(1))
             else "non-descript")

    ftype = normalise_string(ftype).lower()

    accum = []
    for line in block:
        numbers = get_numbers(line)
        if "*  x" in line:
            accum += numbers[0:]
        elif "*  y" in line:
            accum += numbers[1:]
        elif "*  z" in line:
            accum += numbers[2:]

    accum = to_type(accum, float)

    return ftype, accum


def _process_initial_spins(block: TextIO) -> Dict[AtomIndex, Union[float, bool]]:

    accum = {}
    for line in block:
        if match := re.match(rf"\s*\|\s*{REs.ATREG}\s*"
                             rf"{labelled_floats(('spin', 'magmom'))}\s*"
                             r"(?P<fix>[TF])\s*\|", line):
            match = match.groupdict()
            ind = atreg_to_index(match)
            fix_data_types(match, {"spin": float, "magmom": float})
            match["fix"] = match["fix"] == "T"
            accum[ind] = match
    return accum


def _process_born(block: TextIO) -> Dict[AtomIndex, ThreeByThreeMatrix]:
    """ Process a Born block into a dict of charges """

    born_accum = {}
    for line in block:
        if match := REs.BORN_RE.match(line):
            born_accum[atreg_to_index(match)] = (to_type(match["charges"].split(), float),
                                                 to_type(next(block).split(), float),
                                                 to_type(next(block).split(), float))
    return born_accum


def _process_raman(block: TextIO) -> List[Dict[str, Union[ThreeVector, ThreeByThreeMatrix]]]:
    """ Process a Mulliken block into a list of modes """

    next(block)  # Skip first captured line
    modes = []
    curr_mode = {}
    for line in block:
        if "Mode number" in line:
            if curr_mode:
                modes.append(curr_mode)
            curr_mode = {"tensor": [], "depolarisation": None}
        elif numbers := get_numbers(line):
            curr_mode["tensor"].append(to_type(numbers[0:3], float))
            if len(numbers) == 4:
                curr_mode["depolarisation"] = to_type(numbers[3], float)

        elif re.search(r"^ \+\s+\+", line):  # End of 3x3+depol block
            # Compute Invariants Tr(A) and Tr(A)^2-Tr(A^2) of Raman Tensor
            curr_mode["tensor"] = tuple(curr_mode["tensor"])
            tensor = curr_mode["tensor"]
            curr_mode["trace"] = sum(tensor[i][i] for i in range(3))
            curr_mode["II"] = (tensor[0][0]*tensor[1][1] +
                               tensor[0][0]*tensor[2][2] +
                               tensor[1][1]*tensor[2][2] -
                               tensor[0][1]*tensor[1][0] -
                               tensor[0][2]*tensor[2][0] -
                               tensor[1][2]*tensor[2][1])
    if curr_mode:
        modes.append(curr_mode)

    return modes


def _process_mulliken(block: TextIO) -> Dict[AtomIndex, Dict[str, float]]:
    """ Process a mulliken block into a dict of points """
    accum = {}

    for line in block:
        if match := REs.POPN_RE.match(line):
            mull = match.groupdict()
            mull["spin_sep"] = bool(mull["spin_sep"])
            if mull["spin_sep"]:  # We have spin separation
                add_aliases(mull,
                            {orb: f"up_{orb}" for orb in (*SHELLS, "total")},
                            replace=True)
                line = next(block)
                match = REs.POPN_RE_DN.match(line)
                match = match.groupdict()

                add_aliases(match,
                            {orb: f"dn_{orb}" for orb in (*SHELLS, "total")},
                            replace=True)

                mull.update(match)
                mull["total"] = float(mull["up_total"]) + float(mull["dn_total"])

            ind = atreg_to_index(mull)
            fix_data_types(mull, {**{f"{orb}": float for orb in (*SHELLS, "total",
                                                                 "charge", "spin")},
                                  **{f"up_{orb}": float for orb in (*SHELLS, "total")},
                                  **{f"dn_{orb}": float for orb in (*SHELLS, "total")}})
            accum[ind] = mull

    return accum


def _process_band_structure(block: TextIO) -> List[Dict[str, Union[float, int]]]:
    """ Process a band structure into a list of kpts"""

    def fdt(qdat):
        fix_data_types(qdat, {"spin": int,
                              "kx": float,
                              "ky": float,
                              "kz": float,
                              "kpgrp": int,
                              "band": int,
                              "energy": float})

    bands = []
    qdata = {}

    for line in block:
        if match := REs.BS_RE.search(line):
            if qdata:
                fdt(qdata)
                bands.append(qdata)
            qdata = {"band": [], "energy": [], **match.groupdict()}

        elif match := re.search(labelled_floats(("band", "energy"), sep=r"\s+"), line):
            stack_dict(qdata, match.groupdict())

    if qdata:
        fdt(qdata)
        bands.append(qdata)

    return bands


def _process_qdata(qdata: Dict[str, str]) -> Dict[str, Union[float, int]]:
    """ Special parse for phonon qdata """
    qdata = {key: val
             for key, val in qdata.items()
             if any(val) or key == "qpt"}
    fix_data_types(qdata,
                   {"qpt": float,
                    "N": int,
                    "frequency": float,
                    "intensity": float,
                    "raman_intensity": float
                    })
    return qdata


def _parse_magres_block(task: int, inp: TextIO) -> Dict[str, Dict[AtomIndex, float]]:
    """ Parse MagRes data tables from inp according to task """

    data = defaultdict(list)
    data["task"] = REs.MAGRES_TASK[task]
    curr_re = REs.MAGRES_RE[task]
    for line in inp:
        if match := curr_re.match(line):
            match = match.groupdict()
            ind = atreg_to_index(match)
            fix_data_types(match, {key: float for key in ("iso", "aniso", "cq", "eta",
                                                          "fc", "sd", "para", "dia", "tot")})

            if "asym" in match:
                match["asym"] = float(match["asym"]) if match["asym"] != "N/A" else None

            data[ind] = match

    return data


def _process_finalisation(block: TextIO) -> Dict[str, float]:

    out = {}

    for line in block:
        if line.strip():
            key, val = line.split("=")
            out[normalise_string(key.lower())] = to_type(get_numbers(val)[0], float)
    return out


def _process_memory_est(block: TextIO) -> Dict[str, float]:

    accum = {}

    for line in block:
        if match := re.match(r"\s*\|([A-Za-z ]+)" +
                             labelled_floats(("memory", "disk"), suff=" MB"), line):
            key, memory, disk = match.groups()
            accum[normalise_string(key)] = {"memory": float(memory),
                                            "disk": float(disk)}

    return accum


def _process_phonon_sym_analysis(block: TextIO) -> Dict[str, Union[str, List[List[float]]]]:

    accum = {}
    accum["title"] = normalise_string(next(block).split(":")[1])
    next(block)
    accum["mat"] = [to_type(numbers, int) if all(map(lambda x: x.isdigit(), numbers))
                    else to_type(numbers, float)
                    for line in block if (numbers := get_numbers(line))]
    return accum


def _process_kpoint_blocks(block: TextIO,
                           explicit_kpoints: bool) -> Dict[str, Union[ThreeVector, float, int]]:

    if explicit_kpoints:
        accum = {}
        for line in block:
            if "MP grid size" in line:
                accum["kpoint_mp_grid"] = to_type(get_numbers(line), int)
            elif "offset" in line:
                accum["kpoint_mp_offset"] = to_type(get_numbers(line), float)
            elif "Number of kpoints" in line:
                accum["num_kpoints"] = to_type(get_numbers(line)[0], int)
    else:

        accum = {'points': [{"qpt": to_type(match.group("qx", "qy", "qx"), float),
                             "weight": to_type(match["wt"], float)}
                            for line in block
                            if (match :=
                                re.match(
                                    gen_table_re(r"\d\s*" +
                                                 labelled_floats(('qx', 'qy', 'qz', 'wt')), r"\+"),
                                    line))]}
        accum["num_kpoints"] = len(accum["points"])

    return accum


def _process_symmetry(block: TextIO) -> Dict[str, Any]:

    sym = {}
    con = {}

    for line in block:
        if "=" in line:
            key, val = map(normalise_string, line.split("="))

            if "Number of" in line:
                val = to_type(val, int)

            if "constraints" in key:
                con[key] = val
            else:
                sym[key] = val

        elif re.match(r"\s*\d+\s*rotation\s*", line):
            if "symop" not in sym:
                sym["symop"] = []

            curr_sym = {"rotation": [], "symmetry_related": []}
            for curr_ln in itertools.islice(block, 3):
                curr_sym["rotation"].append(to_type(get_numbers(curr_ln), float))

            next(block)  # elif re.match(r"\s*\d+\s*displacement\s*", line):

            curr_sym["displacement"] = to_type(get_numbers(next(block)), float)

            next(block)  # elif "symmetry related atoms:" in line:

            while line := next(block).strip():
                key, val = line.split(":")
                curr_sym["symmetry_related"].extend((key, int(ind))
                                                    for ind in val.split())

            sym['symop'].append(curr_sym)

        elif "Centre of mass" in line:
            con["com_constrained"] = "NOT" not in line

        elif cons_block := get_block(line, block,
                                     r"constraints\.{5}", r"\s*x+\.{4}\s*", out_fmt=str):
            con["ionic_constraints"] = defaultdict(list)
            for match in re.finditer(rf"{REs.ATREG}\s*[xyz]\s*" +
                                     labelled_floats(('pos',), counts=(3,)),
                                     cons_block):
                match = match.groupdict()
                ind = atreg_to_index(match)
                con["ionic_constraints"][ind].append(to_type(match["pos"].split(), float))
        elif "Cell constraints are:" in line:
            con["cell_constraints"] = to_type(get_numbers(line), int)

    return sym, con


def _process_dynamical_matrix(block: TextIO) -> Tuple[complex]:
    next(block)  # Skip header line
    next(block)

    real_part = []
    for line in block:
        if "Ion" in line:
            break
        numbers = get_numbers(line)
        real_part.append(numbers[2:])

    imag_part = []
    # Get remainder
    for line in block:
        if numbers := get_numbers(line):
            imag_part.append(numbers[2:])

    accum = []
    for real_row, imag_row in zip(real_part, imag_part):
        accum.append(tuple(complex(float(real), float(imag))
                           for real, imag in zip(real_row, imag_row)))

    return tuple(accum)


def _process_pspot_string(string: str) -> Dict[str, Union[float, int, str]]:
    if not (match := REs.PSPOT_RE.match(string)):
        raise IOError(f"Attempt to parse {string} as PSPot failed")

    pspot = match.groupdict()
    projectors = []
    for proj in pspot["proj"].split(":"):
        pdict = REs.PSPOT_PROJ_RE.match(proj).groupdict()
        pdict["shell"] = SHELLS[int(pdict["shell"])]
        fix_data_types(pdict, {"orbital": int})
        projectors.append(pdict)

    if not pspot["shell_swp"]:
        del pspot["shell_swp"]

    if not pspot["shell_swp2"]:
        del pspot["shell_swp2"]

    pspot["projectors"] = tuple(projectors)
    pspot["string"] = string

    fix_data_types(pspot, {"beta_radius": float,
                           "r_inner": float,
                           "core_radius": float,
                           "coarse": int,
                           "medium": int,
                           "fine": int,
                           "local_channel": int})

    return pspot


def _process_pspot_report(block: TextIO) -> Dict[str, Union[float, str]]:

    accum = {"reference_electronic_structure": [],
             "pseudopotential_definition": []}

    for line in block:
        if match := REs.PSPOT_REFERENCE_STRUC_RE.match(line):
            match = match.groupdict()
            fix_data_types(match, {"occupation": float, "energy": float})
            accum["reference_electronic_structure"].append(match)
        elif match := REs.PSPOT_DEF_RE.match(line):
            match = match.groupdict()
            # Account for "loc"
            match["beta"] = int(match["beta"]) if match["beta"].isnumeric() else match["beta"]
            fix_data_types(match, {"l": int, "j": int,
                                   "e": float, "Rc": float, "norm": int})
            accum["pseudopotential_definition"].append(match)
        elif match := re.search(rf"Element: (?P<element>{REs.SPECIES_RE})\s+"
                                rf"Ionic charge: (?P<ionic_charge>{REs.FNUMBER_RE})\s+"
                                r"Level of theory: (?P<level_of_theory>[\w\d]+)", line):
            match = match.groupdict()
            match["ionic_charge"] = float(match["ionic_charge"])
            accum.update(match)

        elif match := re.search(r"Atomic Solver:\s*(?P<solver>[\w\s-]+)", line):
            accum["solver"] = normalise_string(match["solver"])

        elif match := REs.PSPOT_RE.search(line):
            accum["detail"] = _process_pspot_string(match.group(0))

        elif "Augmentation charge Rinner" in line:
            accum["augmentation_charge_rinner"] = to_type(get_numbers(line), float)

        elif "Partial core correction Rc" in line:
            accum["partial_core_correction"] = to_type(get_numbers(line), float)

    return accum


def _process_bond_analysis(block: TextIO) -> Dict[Tuple[AtomIndex, AtomIndex], Dict[str, float]]:
    accum = {((match["spec1"], int(match["ind1"])),
              (match["spec2"], int(match["ind2"]))): {"population": float(match["population"]),
                                                      "length": float(match["length"])}
             for line in block
             if (match := REs.BOND_RE.match(line))}
    return accum


def _process_orbital_populations(block: TextIO) -> Dict[str, Union[Dict[str, float],
                                                                   float,
                                                                   Tuple[float]]]:

    accum = defaultdict(dict)
    for line in block:
        if match := REs.ORBITAL_POPN_RE.match(line):
            ind = match.groupdict()
            ind = atreg_to_index(ind)
            accum[ind][match["orb"]] = to_type(match["charge"], float)
        elif match := re.match(rf"\s*Total:\s*{labelled_floats(('charge',))}", line):
            accum["total"] = float(match["charge"])
        elif "total projected" in line:
            accum["total_projected"] = to_type(get_numbers(line), float)

    return accum


def _process_dftd(block: TextIO) -> Dict[str, Union[Dict[str, float], float]]:
    dftd = {"species": {}}

    for line in block:
        if len(match := line.split(":")) == 2:
            key, val = match
            val = normalise_string(val)
            if "Parameter" in key:
                val = to_type(get_numbers(val)[0], float)
            dftd[normalise_string(key).lower()] = val

        elif match := re.match(rf"\s*x\s*(?P<spec>{REs.ATOM_NAME_RE})\s*" +
                               labelled_floats(('C6', 'R0')), line):
            dftd["species"][match["spec"]] = {"C6": float(match["C6"]),
                                              "R0": float(match["R0"])}

    return dftd


def _process_occupancies(block: TextIO) -> Dict[str, Union[int, float]]:
    label = ("band", "eigenvalue", "occupancy")

    accum = [dict(zip(label, numbers)) for line in block if (numbers := get_numbers(line))]
    for elem in accum:
        fix_data_types(elem, {"band": int,
                              "eigenvalue": float,
                              "occupancy": float})
    return accum


def _process_wvfn_line_min(block: TextIO) -> Dict[str, Tuple[float]]:
    accum = {}
    for line in block:
        if "initial" in line:
            accum["init_energy"], accum["init_de_dstep"] = to_type(get_numbers(line), float)
        elif line.strip().startswith("| step"):
            accum["steps"] = to_type(get_numbers(line), float)
        elif line.strip().startswith("| gain"):
            accum["gain"] = to_type(get_numbers(line), float)

    return accum


def _process_autosolvation(block: TextIO) -> Dict[str, float]:

    accum = {}
    for line in block:
        if len(match := line.split("=")) > 1:
            key = normalise_string(match[0].strip("-( "))
            val = to_type(get_numbers(line)[0], float)
            accum[key] = val

    return accum


# def _process_phonon(block: TextIO):
#     ...

def _process_dipole(block: TextIO) -> Dict[str, Union[ThreeVector, float]]:

    accum = {}

    for line in block:
        if match := re.search(r"Total\s*(?P<type>\w+)", line):
            accum[f"total_{match['type']}"] = float(get_numbers(line)[0])

        elif "Centre" in line:
            key = "centre_electronic" if "elec" in line else "centre_positive"
            accum[key] = to_type(get_numbers(next(block)), float)

        elif "Magnitude" in line:
            accum["dipole_magnitude"] = to_type(get_numbers(line)[0], float)

        elif "Direction" in line:
            accum["dipole_direction"] = to_type(get_numbers(line), float)

    return accum


def _process_pair_params(block_in: TextIO) -> Dict[Union[str, Tuple[str]], float]:

    accum = {}
    for line in block_in:
        # Two-body
        if block := get_block(line, block_in, "Two Body", r"^\w*\s*\*+\s*$"):
            for blk_line in block:
                if REs.PAIR_POT_RES['two_body_spec'].search(blk_line):
                    match = REs.PAIR_POT_RES['two_body_spec'].finditer(blk_line)
                    labels = tuple(mch.groups() for mch in match)

                elif match := REs.PAIR_POT_RES['two_body_val'].match(blk_line):
                    tag, typ, lab = match.group("tag", "type", "label")
                    if tag:
                        typ = f"{tag}_{typ}"
                    if typ not in accum:
                        accum[typ] = {}
                    if lab not in accum[typ]:
                        accum[typ][lab] = {}

                    accum[typ][lab].update(zip(labels,
                                               to_type(match["params"].split(),
                                                       float)))

                elif match := REs.PAIR_POT_RES['two_body_one_spec'].match(blk_line):
                    labels = (match["spec"],)

        # Three-body
        elif block := get_block(line, block_in, "Three Body", r"^\s*\*+\s*$"):
            for blk_line in block:
                if match := REs.PAIR_POT_RES['three_body_spec'].match(blk_line):
                    labels = (tuple(match["spec"].split()),)

                elif match := REs.PAIR_POT_RES["three_body_val"].match(blk_line):
                    tag, typ, lab = match.group("tag", "type", "label")

                    if tag:
                        typ = f"{tag}_{typ}"
                    if typ not in accum:
                        accum[typ] = {}
                    if lab not in accum[typ]:
                        accum[typ][lab] = {}

                    accum[typ][lab].update(zip(labels,
                                               to_type(match["params"].split(),
                                                       float)))

        # Globals
        elif match := REs.PAIR_POT_RES["three_body_val"].match(line):
            tag, typ, lab = match.group("tag", "type", "label")

            if tag:
                typ = f"{tag}_{typ}"
            if typ not in accum:
                accum[typ] = {}

            accum[typ][lab] = to_type(match["params"], float)

    return accum


def _process_geom_table(block: TextIO) -> Dict[str, Union[bool, float]]:

    accum = {}
    for line in block:
        if match := REs.GEOMOPT_MIN_TABLE_RE.match(line):
            match = match.groupdict()
            fix_data_types(match, {key: float for key in ('lambda', 'Fdelta', 'enthalpy')})

            key = normalise_string(match["step"])
            del match["step"]
            accum[key] = match

        elif match := REs.GEOMOPT_TABLE_RE.match(line):
            match = match.groupdict()
            fix_data_types(match, {key: float for key in ('value', 'tolerance')})

            match["converged"] = match["converged"] == "Yes"

            key = normalise_string(match["parameter"])
            del match["parameter"]
            accum[key] = match

    return accum


def _process_elastic_properties(block: TextIO) -> Dict[str, Tuple[float]]:
    accum = {}

    for line in block:
        if "::" in line:
            key = line.split("::")[0]
            val = to_type(get_numbers(line), float)

            if len(val) == 1:
                val = val[0]

            accum[normalise_string(key)] = val
        elif blk := get_block(line, block, "Speed of Sound", REs.EMPTY):

            accum["Speed of Sound"] = [to_type(numbers, float)
                                       for blk_line in blk
                                       if (numbers := get_numbers(blk_line))]

    return accum



"""
Setuptools setup script for castep_outputs

To install locally:

    python setup.py install

"""

import sys
from setuptools import setup, find_packages

# Check for valid Python version
if sys.version_info[:2] < (3, 8):
    print(f"castep_outputs requires Python 3.8. Python {'.'.join(sys.version_info[:2])} detected")

packages_test = find_packages()

with open("README.rst", "r", encoding="utf-8") as readme:
    README = readme.read()

setup(
    name="castep_outputs",
    version="0.1.4",
    license='BSD3',
    license_files=('LICENSE'),
    description='A package for extracting information from castep outputs',
    long_description=README,
    packages=find_packages(),
    author="Jacob Wilkins",
    author_email="jacob.wilkins@stfc.ac.uk",
    url="https://github.com/oerc0122/castep_outputs",
    download_url="https://github.com/oerc0122/castep_outputs",
    python_requires='>=3.8',
    install_requires=[],
    extras_require={'ruamel': 'ruamel.yaml>=0.17.22',
                    'yaml': 'PyYAML>=3.13'},
    entry_points={"console_scripts": ['castep_outputs = castep_outputs.castep_outputs_main:main']},
    include_package_data=True
)



"""
Run main castep parser
"""
from typing import Callable, Optional, TextIO, Union
from pathlib import Path
import logging
import fileinput
import io
import sys

from .args import (parse_args, args_to_dict)
from .utility import (json_safe, flatten_dict, get_dumpers)
from .castep_file_parser import parse_castep_file
from .cell_param_file_parser import parse_cell_param_file
from .md_geom_file_parser import parse_md_geom_file
from .extra_files_parser import (parse_bands_file, parse_hug_file, parse_phonon_dos_file,
                                 parse_efield_file, parse_xrd_sf_file, parse_elf_fmt_file,
                                 parse_chdiff_fmt_file, parse_pot_fmt_file, parse_den_fmt_file,
                                 parse_elastic_file, parse_ts_file)
from .constants import OutFormats

PARSERS = {
    ".castep": parse_castep_file,
    ".cell": parse_cell_param_file,
    ".param": parse_cell_param_file,
    ".geom": parse_md_geom_file,
    ".md": parse_md_geom_file,
    ".bands": parse_bands_file,
    ".hug": parse_hug_file,
    ".phonon_dos": parse_phonon_dos_file,
    ".efield": parse_efield_file,
    ".xrd_sf": parse_xrd_sf_file,
    ".elf_fmt": parse_elf_fmt_file,
    ".chdiff_fmt": parse_chdiff_fmt_file,
    ".pot_fmt": parse_pot_fmt_file,
    ".den_fmt": parse_den_fmt_file,
    ".elastic": parse_elastic_file,
    ".ts": parse_ts_file
    }


def parse_single(in_file: Union[Path, TextIO], parser: Callable = None,
                 out_format: OutFormats = "print",
                 *, loglevel: int = logging.WARNING, testing: bool = False):
    """
    Parse a file using the given parser and post-process according to options
    """

    logging.basicConfig(format="%(levelname)s: %(message)s", level=loglevel)

    if not isinstance(in_file, (io.TextIOBase, Path)):
        in_file = Path(in_file)

    if parser is None:
        ext = in_file.suffix

        if ext not in PARSERS:
            raise KeyError(f"Parser for file {in_file} (assumed type: {ext}) not found")

        parser = PARSERS[ext]

    if isinstance(in_file, io.TextIOBase):
        data = parser(in_file)
    else:
        with fileinput.FileInput(in_file, mode='r', encoding='utf-8') as file:
            data = parser(file)

    if out_format == "json" or testing:
        data = json_safe(data)

    if testing:
        if isinstance(data, list):
            data = [flatten_dict(run) for run in data]
        else:
            data = flatten_dict(data)

    return data


def parse_all(output: Optional[Path] = None, out_format: OutFormats = "json",
              *, loglevel: int = logging.WARNING, testing: bool = False, **files):
    """ Parse all files in files dict """
    file_dumper = get_dumpers(out_format)

    for typ, paths in files.items():
        parser = PARSERS[f".{typ}"]
        for path in paths:
            data = parse_single(path, parser, out_format, loglevel=loglevel, testing=testing)

            if output is None:
                file_dumper(data, sys.stdout)
                print()
            elif isinstance(output, io.TextIOBase):
                file_dumper(data, output)
            else:
                with open(output, 'a+', encoding='utf-8') as out_file:
                    file_dumper(data, out_file)


def main():
    """ Run the main program from command line """
    args = parse_args()
    dict_args = args_to_dict(args)

    parse_all(output=args.output,
              loglevel=getattr(logging, args.log.upper()),
              testing=args.testing,
              out_format=args.out_format,
              **dict_args)


if __name__ == "__main__":
    main()
